@@ -20,13 +20,10 @@
 #include <gtk/gtk.h>
 #include <libaudcore/i18n.h>
 #include <libaudgui/libaudgui.h>
+#include <libaudgui/libaudgui-gtk.h>
 
 #include "callbacks.h"
 
-#if GTK_CHECK_VERSION (3, 12, 0)
-#define gtk_widget_set_margin_right gtk_widget_set_margin_end
-#endif
-
 static const char help_text[] =
  N_("Time\n"
     "  Alarm at:\n"
@@ -106,13 +103,6 @@ GtkWidget *create_reminder_dialog (const char *reminder_msg)
     return reminder_dialog;
 }
 
-static void file_set_cb (GtkFileChooserButton *button, void * entry)
-{
-    char *uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (button));
-    gtk_entry_set_text (GTK_ENTRY (entry), uri);
-    g_free (uri);
-}
-
 GtkWidget *create_config_notebook ()
 {
     /* General */
@@ -154,8 +144,8 @@ GtkWidget *create_config_notebook ()
     GtkWidget *fading_spin, *quiet_vol_scale, *vol_scale, *separator, *current_button;
 
     /* Page 4 */
-    GtkWidget  *cmd_entry, *playlist_entry, *reminder_text;
-    GtkWidget *cmd_checkb, *reminder_checkb, *file_chooser_button;
+    GtkWidget *cmd_entry, *playlist_entry, *reminder_text;
+    GtkWidget *cmd_checkb, *reminder_checkb;
 
     /* Page 5 */
     GtkWidget *view, *scrolled_window;
@@ -180,6 +170,7 @@ GtkWidget *create_config_notebook ()
 
     adjustment = gtk_adjustment_new (6, 0, 23, 1, 10, 0);
     alarm_h_spin = gtk_spin_button_new (adjustment, 1, 0);
+    gtk_widget_set_valign (alarm_h_spin, GTK_ALIGN_CENTER);
     g_object_set_data (G_OBJECT (notebook), "alarm_h_spin", alarm_h_spin);
     gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (alarm_h_spin), GTK_UPDATE_IF_VALID);
     gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (alarm_h_spin), true);
@@ -191,6 +182,7 @@ GtkWidget *create_config_notebook ()
 
     adjustment = gtk_adjustment_new (30, 0, 59, 1, 10, 0);
     alarm_m_spin = gtk_spin_button_new (adjustment, 1, 0);
+    gtk_widget_set_valign (alarm_m_spin, GTK_ALIGN_CENTER);
     g_object_set_data (G_OBJECT (notebook), "alarm_m_spin", alarm_m_spin);
     gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (alarm_m_spin), GTK_UPDATE_IF_VALID);
     gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (alarm_m_spin), true);
@@ -209,6 +201,7 @@ GtkWidget *create_config_notebook ()
 
     adjustment = gtk_adjustment_new (0, 0, 100, 1, 10, 0);
     stop_h_spin = gtk_spin_button_new (adjustment, 1, 0);
+    gtk_widget_set_valign (stop_h_spin, GTK_ALIGN_CENTER);
     g_object_set_data (G_OBJECT (notebook), "stop_h_spin", stop_h_spin);
     gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (stop_h_spin), GTK_UPDATE_IF_VALID);
     gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (stop_h_spin), true);
@@ -221,6 +214,7 @@ GtkWidget *create_config_notebook ()
 
     adjustment = gtk_adjustment_new (0, 0, 59, 1, 10, 0);
     stop_m_spin = gtk_spin_button_new (adjustment, 1, 0);
+    gtk_widget_set_valign (stop_m_spin, GTK_ALIGN_CENTER);
     g_object_set_data (G_OBJECT (notebook), "stop_m_spin", stop_m_spin);
     gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (stop_m_spin), GTK_UPDATE_IF_VALID);
     gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (stop_m_spin), true);
@@ -385,14 +379,10 @@ GtkWidget *create_config_notebook ()
     hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 5);
     gtk_container_set_border_width (GTK_CONTAINER (frame), 10);
     gtk_container_set_border_width (GTK_CONTAINER (hbox), 8);
-    playlist_entry = gtk_entry_new ();
+    playlist_entry = audgui_file_entry_new (GTK_FILE_CHOOSER_ACTION_OPEN, _("Select a playlist"));
     g_object_set_data (G_OBJECT (notebook), "playlist", playlist_entry);
 
-    file_chooser_button = gtk_file_chooser_button_new (_("Select a playlist"), GTK_FILE_CHOOSER_ACTION_OPEN);
-    gtk_widget_set_valign (file_chooser_button, GTK_ALIGN_CENTER);
-    g_signal_connect (file_chooser_button, "file-set", G_CALLBACK (file_set_cb), playlist_entry);
     gtk_box_pack_start (GTK_BOX (hbox), playlist_entry, true, true, 0);
-    gtk_container_add (GTK_CONTAINER (hbox), file_chooser_button);
     gtk_container_add (GTK_CONTAINER (frame), hbox);
     gtk_container_add (GTK_CONTAINER (vbox), frame);
 
