@@ -101,17 +101,16 @@ static void
 aosd_cb_configure_position_placement_commit ( GtkWidget * grid , aosd_cfg_t * cfg )
 {
   GList *placbt_list = gtk_container_get_children( GTK_CONTAINER(grid) );
-  GList *list_iter = placbt_list;
 
-  while ( list_iter != nullptr )
+  for ( GList *iter = placbt_list; iter != nullptr; iter = iter->next )
   {
-    GtkWidget *placbt = (GtkWidget *) list_iter->data;
+    GtkWidget *placbt = (GtkWidget *) iter->data;
+
     if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(placbt) ) == true )
     {
       cfg->position.placement = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(placbt),"value"));
       break;
     }
-    list_iter = g_list_next( list_iter );
   }
 
   g_list_free( placbt_list );
@@ -643,7 +642,6 @@ aosd_ui_configure_trigger ( aosd_cfg_t * cfg )
 }
 
 
-#ifdef HAVE_XCOMPOSITE
 static void
 aosd_cb_configure_misc_transp_real_clicked ( GtkToggleButton * real_rbt , void * status_hbox )
 {
@@ -673,22 +671,23 @@ aosd_cb_configure_misc_transp_real_clicked ( GtkToggleButton * real_rbt , void *
     gtk_widget_set_sensitive( GTK_WIDGET(status_hbox) , false );
   }
 }
-#endif
 
 
 static void
 aosd_cb_configure_misc_transp_commit ( GtkWidget * mis_transp_vbox , aosd_cfg_t * cfg )
 {
   GList *child_list = gtk_container_get_children( GTK_CONTAINER(mis_transp_vbox) );
-  while (child_list != nullptr)
+
+  for ( GList *iter = child_list; iter != nullptr; iter = iter->next )
   {
-    if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(child_list->data) ) )
+    if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(iter->data) ) )
     {
-      cfg->misc.transparency_mode = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(child_list->data),"val"));
+      cfg->misc.transparency_mode = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(iter->data), "val"));
       break;
     }
-    child_list = g_list_next(child_list);
   }
+
+  g_list_free( child_list );
 }
 
 
@@ -736,7 +735,6 @@ aosd_ui_configure_misc ( aosd_cfg_t * cfg )
   g_object_set_data( G_OBJECT(mis_transp_status_hbox) , "img" , mis_transp_status_img );
   g_object_set_data( G_OBJECT(mis_transp_status_hbox) , "label" , mis_transp_status_label );
 
-#ifdef HAVE_XCOMPOSITE
   g_signal_connect( G_OBJECT(mis_transp_real_rbt) , "toggled" ,
     G_CALLBACK(aosd_cb_configure_misc_transp_real_clicked) , mis_transp_status_hbox );
 
@@ -757,14 +755,6 @@ aosd_ui_configure_misc ( aosd_cfg_t * cfg )
     gtk_label_set_text( GTK_LABEL(mis_transp_status_label) , _("Composite extension not loaded") );
     gtk_widget_set_sensitive( GTK_WIDGET(mis_transp_status_hbox) , false );
   }
-#else
-  gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(mis_transp_fake_rbt) , true );
-  gtk_widget_set_sensitive( GTK_WIDGET(mis_transp_real_rbt) , false );
-  gtk_image_set_from_icon_name( GTK_IMAGE(mis_transp_status_img) ,
-    "dialog-error" , GTK_ICON_SIZE_MENU );
-  gtk_label_set_text( GTK_LABEL(mis_transp_status_label) , _("Composite extension not available") );
-  gtk_widget_set_sensitive( GTK_WIDGET(mis_transp_status_hbox) , false );
-#endif
 
   aosd_cb_list.append( mis_transp_vbox , aosd_cb_configure_misc_transp_commit );
 
