@@ -22,6 +22,7 @@
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
+#define AUD_GLIB_INTEGRATION
 #include <libaudcore/i18n.h>
 #include <libaudcore/runtime.h>
 #include <libaudcore/plugins.h>
@@ -196,9 +197,8 @@ static GtkWidget * vbox_new (GtkWidget * widget, const char * name)
      widget);
 
     GtkWidget * label = gtk_label_new (nullptr);
-    char * markup = g_markup_printf_escaped ("<small><b>%s</b></small>", name);
+    CharPtr markup (g_markup_printf_escaped ("<small><b>%s</b></small>", name));
     gtk_label_set_markup ((GtkLabel *) label, markup);
-    g_free (markup);
     gtk_widget_set_halign (label, GTK_ALIGN_START);
     gtk_container_add ((GtkContainer *) ebox, label);
 
@@ -567,14 +567,17 @@ void layout_save ()
         Item * item = (Item *) node->data;
         g_return_if_fail (item && item->name);
 
-        char key[16], value[64];
+        char key[32], value[64];
 
         snprintf (key, sizeof key, "item%d_name", i);
         aud_set_str ("gtkui-layout", key, item->name);
 
+        int temp_w = audgui_to_portable_dpi (item->w);
+        int temp_h = audgui_to_portable_dpi (item->h);
+
         snprintf (key, sizeof key, "item%d_pos", i);
         snprintf (value, sizeof value, "%d,%d,%d,%d,%d", item->dock, item->x,
-         item->y, item->w, item->h);
+         item->y, temp_w, temp_h);
         aud_set_str ("gtkui-layout", key, value);
 
         i ++;
@@ -591,15 +594,22 @@ void layout_load ()
 
     for (int i = 0; i < count; i ++)
     {
-        char key[16];
+        char key[32];
 
         snprintf (key, sizeof key, "item%d_name", i);
         String name = aud_get_str ("gtkui-layout", key);
         Item * item = item_new (name);
 
+        int temp_w = 0, temp_h = 0;
+
         snprintf (key, sizeof key, "item%d_pos", i);
         String pos = aud_get_str ("gtkui-layout", key);
-        sscanf (pos, "%d,%d,%d,%d,%d", & item->dock, & item->x, & item->y, & item->w, & item->h);
+        sscanf (pos, "%d,%d,%d,%d,%d", & item->dock, & item->x, & item->y, & temp_w, & temp_h);
+
+        if (temp_w)
+            item->w = audgui_to_native_dpi (temp_w);
+        if (temp_h)
+            item->h = audgui_to_native_dpi (temp_h);
     }
 }
 
