@@ -22,7 +22,6 @@
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
-#define AUD_PLUGIN_GLIB_ONLY
 #include <libaudcore/audstrings.h>
 #include <libaudcore/drct.h>
 #include <libaudcore/hook.h>
@@ -41,7 +40,8 @@
 #include "ui_playlist_widget.h"
 #include "ui_infoarea.h"
 #include "ui_statusbar.h"
-#include "playlist_util.h"
+
+#include "../ui-common/menu-ops.h"
 
 static const char * const gtkui_defaults[] = {
     "infoarea_show_vis", "TRUE",
@@ -55,12 +55,13 @@ static const char * const gtkui_defaults[] = {
     "autoscroll", "TRUE",
     "playlist_columns", "title artist album queued length",
     "playlist_headers", "TRUE",
-    "record", "FALSE",
     "show_remaining_time", "FALSE",
     "step_size", "5",
 
     "player_x", "-1000",
     "player_y", "-1000",
+    "player_width", "768",
+    "player_height", "480",
 
     nullptr
 };
@@ -72,7 +73,8 @@ public:
         N_("GTK Interface"),
         PACKAGE,
         nullptr,
-        & gtkui_prefs
+        & gtkui_prefs,
+        PluginGLibOnly
     };
 
     constexpr GtkUI () : IfacePlugin (info) {}
@@ -106,6 +108,9 @@ public:
         { audgui_plugin_menu_add (id, func, name, icon); }
     void plugin_menu_remove (AudMenuID id, void func ())
         { audgui_plugin_menu_remove (id, func); }
+
+    void startup_notify (const char * id)
+        { gdk_notify_startup_complete_with_id (id); }
 };
 
 EXPORT GtkUI aud_plugin_instance;
@@ -137,22 +142,16 @@ static void save_window_size ()
 
     aud_set_int ("gtkui", "player_x", x);
     aud_set_int ("gtkui", "player_y", y);
-    aud_set_int ("gtkui", "player_width", w);
-    aud_set_int ("gtkui", "player_height", h);
+    aud_set_int ("gtkui", "player_width", audgui_to_portable_dpi (w));
+    aud_set_int ("gtkui", "player_height", audgui_to_portable_dpi (h));
 }
 
 static void restore_window_size ()
 {
-    int dpi = audgui_get_dpi ();
     int x = aud_get_int ("gtkui", "player_x");
     int y = aud_get_int ("gtkui", "player_y");
-    int w = aud_get_int ("gtkui", "player_width");
-    int h = aud_get_int ("gtkui", "player_height");
-
-    if (w < 1)
-        w = 8 * dpi;
-    if (h < 1)
-        h = 5 * dpi;
+    int w = audgui_to_native_dpi (aud_get_int ("gtkui", "player_width"));
+    int h = audgui_to_native_dpi (aud_get_int ("gtkui", "player_height"));
 
     gtk_window_set_default_size ((GtkWindow *) window, w, h);
 
@@ -214,19 +213,23 @@ static void title_change (void * = nullptr)
 {
     delayed_title_change.stop ();
 
+    StringBuf title;
+
     if (aud_drct_get_playing ())
     {
         if (aud_drct_get_ready ())
-        {
-            String title = aud_drct_get_title ();
-            gtk_window_set_title ((GtkWindow *) window,
-             str_printf (_("%s - Audacious"), (const char *) title));
-        }
+            title.steal (str_printf (_("%s - Audacious"), (const char *) aud_drct_get_title ()));
         else
-            gtk_window_set_title ((GtkWindow *) window, _("Buffering ..."));
+            title.steal (str_copy (_("Buffering ...")));
     }
     else
-        gtk_window_set_title ((GtkWindow *) window, _("Audacious"));
+        title.steal (str_copy (_("Audacious")));
+
+    int instance = aud_get_instance ();
+    if (instance != 1)
+        title.combine (str_printf (" (%d)", instance));
+
+    gtk_window_set_title ((GtkWindow *) window, title);
 }
 
 void GtkUI::show (bool show)
@@ -321,12 +324,13 @@ static void do_seek (int time)
 static gboolean ui_slider_change_value_cb (GtkRange * range,
  GtkScrollType scroll, double value)
 {
-    int time = value;
+    int length = aud_drct_get_length ();
+    int time = aud::clamp ((int) value, 0, length);
 
     if (slider_is_moving)
     {
         slider_seek_time = time;
-        set_time_label (time, aud_drct_get_length ());
+        set_time_label (time, length);
     }
     else if (time != slider_seek_time)  // avoid seeking twice
         do_seek (time);
@@ -334,11 +338,20 @@ static gboolean ui_slider_change_value_cb (GtkRange * range,
     return false;
 }
 
+/* return an object property if it exists, otherwise false */
+static bool get_boolean_prop (void * obj, const char * prop)
+{
+    gboolean value = false;
+    if (g_object_class_find_property (G_OBJECT_GET_CLASS (obj), prop))
+        g_object_get (obj, prop, & value, nullptr);
+
+    return value;
+}
+
 static gboolean ui_slider_button_press_cb (GtkWidget * widget, GdkEventButton * event)
 {
-    gboolean primary_warps = false;
-    g_object_get (gtk_widget_get_settings (widget),
-     "gtk-primary-button-warps-slider", & primary_warps, nullptr);
+    bool primary_warps = get_boolean_prop (gtk_widget_get_settings (widget),
+     "gtk-primary-button-warps-slider");
 
     if (event->button == 1 && ! primary_warps)
         event->button = 2;
@@ -349,9 +362,8 @@ static gboolean ui_slider_button_press_cb (GtkWidget * widget, GdkEventButton *
 
 static gboolean ui_slider_button_release_cb (GtkWidget * widget, GdkEventButton * event)
 {
-    gboolean primary_warps = false;
-    g_object_get (gtk_widget_get_settings (widget),
-     "gtk-primary-button-warps-slider", & primary_warps, nullptr);
+    bool primary_warps = get_boolean_prop (gtk_widget_get_settings (widget),
+     "gtk-primary-button-warps-slider");
 
     if (event->button == 1 && ! primary_warps)
         event->button = 2;
@@ -476,11 +488,6 @@ static GtkWidget * markup_label_new (const char * str)
     return label;
 }
 
-static void window_mapped_cb (GtkWidget * widget)
-{
-    gtk_widget_grab_focus (playlist_get_treeview (aud_playlist_get_active ()));
-}
-
 static gboolean window_keypress_cb (GtkWidget * widget, GdkEventKey * event)
 {
     switch (event->state & (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK))
@@ -492,8 +499,8 @@ static gboolean window_keypress_cb (GtkWidget * widget, GdkEventKey * event)
         /* escape key returns focus to playlist */
         if (event->keyval == GDK_KEY_Escape)
         {
-            if (! focused || ! gtk_widget_is_ancestor (focused, (GtkWidget *) UI_PLAYLIST_NOTEBOOK))
-                gtk_widget_grab_focus (playlist_get_treeview (aud_playlist_get_active ()));
+            if (! focused || ! gtk_widget_is_ancestor (focused, pl_notebook))
+                pl_notebook_grab_focus ();
 
             return false;
         }
@@ -538,8 +545,7 @@ static gboolean window_keypress_cb (GtkWidget * widget, GdkEventKey * event)
         {
           case GDK_KEY_ISO_Left_Tab:
           case GDK_KEY_Tab:
-            aud_playlist_set_active ((aud_playlist_get_active () + 1) %
-             aud_playlist_count ());
+            pl_next ();
             break;
 
           default:
@@ -551,8 +557,7 @@ static gboolean window_keypress_cb (GtkWidget * widget, GdkEventKey * event)
         {
           case GDK_KEY_ISO_Left_Tab:
           case GDK_KEY_Tab:
-            aud_playlist_set_active (aud_playlist_get_active () ?
-             aud_playlist_get_active () - 1 : aud_playlist_count () - 1);
+            pl_prev ();
             break;
           default:
             return false;
@@ -587,10 +592,10 @@ static gboolean playlist_keypress_cb (GtkWidget * widget, GdkEventKey * event)
         switch (event->keyval)
         {
         case GDK_KEY_Escape:
-            ui_playlist_notebook_position (aud::to_ptr (aud_playlist_get_active ()), nullptr);
+            pl_notebook_set_position (aud::to_ptr (Playlist::active_playlist ()), nullptr);
             return true;
         case GDK_KEY_Delete:
-            playlist_delete_selected ();
+            pl_remove_selected ();
             return true;
         case GDK_KEY_Menu:
             popup_menu_rclick (0, event->time);
@@ -602,16 +607,16 @@ static gboolean playlist_keypress_cb (GtkWidget * widget, GdkEventKey * event)
         switch (event->keyval)
         {
         case 'x':
-            playlist_cut ();
+            pl_cut ();
             return true;
         case 'c':
-            playlist_copy ();
+            pl_copy ();
             return true;
         case 'v':
-            playlist_paste ();
+            pl_paste ();
             return true;
         case 'a':
-            aud_playlist_select_all (aud_playlist_get_active (), true);
+            pl_select_all ();
             return true;
         }
 
@@ -623,6 +628,10 @@ static gboolean playlist_keypress_cb (GtkWidget * widget, GdkEventKey * event)
 
 static void update_toggles (void * = nullptr, void * = nullptr)
 {
+    gtk_widget_set_visible ((GtkWidget *) button_record, aud_drct_get_record_enabled ());
+
+    gtk_toggle_tool_button_set_active ((GtkToggleToolButton *) button_record,
+     aud_get_bool (nullptr, "record"));
     gtk_toggle_tool_button_set_active ((GtkToggleToolButton *) button_repeat,
      aud_get_bool (nullptr, "repeat"));
     gtk_toggle_tool_button_set_active ((GtkToggleToolButton *) button_shuffle,
@@ -641,24 +650,7 @@ static void toggle_shuffle (GtkToggleToolButton * button)
 
 static void toggle_record (GtkToggleToolButton * button)
 {
-    if (! aud_drct_enable_record (gtk_toggle_tool_button_get_active (button)))
-        gtk_toggle_tool_button_set_active (button, aud_drct_get_record_enabled ());
-}
-
-static void record_toggled (void * = nullptr, void * = nullptr)
-{
-    bool supported = (bool) aud_drct_get_record_plugin ();
-    bool enabled = aud_drct_get_record_enabled ();
-
-    gtk_widget_set_sensitive ((GtkWidget *) button_record, supported);
-    gtk_toggle_tool_button_set_active ((GtkToggleToolButton *) button_record, enabled);
-
-    /* update menu item */
-    if (enabled != aud_get_bool ("gtkui", "record"))
-    {
-        aud_set_bool ("gtkui", "record", enabled);
-        hook_call ("gtkui set record", nullptr);
-    }
+    aud_set_bool (nullptr, "record", gtk_toggle_tool_button_get_active (button));
 }
 
 static void toggle_search_tool (GtkToggleToolButton * button)
@@ -690,13 +682,14 @@ static void ui_hooks_associate ()
     hook_associate ("playback pause", (HookFunction) pause_cb, nullptr);
     hook_associate ("playback unpause", (HookFunction) pause_cb, nullptr);
     hook_associate ("playback stop", (HookFunction) ui_playback_stop, nullptr);
-    hook_associate ("playlist update", ui_playlist_notebook_update, nullptr);
-    hook_associate ("playlist activate", ui_playlist_notebook_activate, nullptr);
-    hook_associate ("playlist set playing", ui_playlist_notebook_set_playing, nullptr);
-    hook_associate ("playlist position", ui_playlist_notebook_position, nullptr);
+    hook_associate ("playlist update", pl_notebook_update, nullptr);
+    hook_associate ("playlist activate", pl_notebook_activate, nullptr);
+    hook_associate ("playlist set playing", pl_notebook_set_playing, nullptr);
+    hook_associate ("playlist position", pl_notebook_set_position, nullptr);
+    hook_associate ("enable record", update_toggles, nullptr);
+    hook_associate ("set record", update_toggles, nullptr);
     hook_associate ("set shuffle", update_toggles, nullptr);
     hook_associate ("set repeat", update_toggles, nullptr);
-    hook_associate ("enable record", record_toggled, nullptr);
     hook_associate ("config save", (HookFunction) config_save, nullptr);
 }
 
@@ -708,13 +701,14 @@ static void ui_hooks_disassociate ()
     hook_dissociate ("playback pause", (HookFunction) pause_cb);
     hook_dissociate ("playback unpause", (HookFunction) pause_cb);
     hook_dissociate ("playback stop", (HookFunction) ui_playback_stop);
-    hook_dissociate ("playlist update", ui_playlist_notebook_update);
-    hook_dissociate ("playlist activate", ui_playlist_notebook_activate);
-    hook_dissociate ("playlist set playing", ui_playlist_notebook_set_playing);
-    hook_dissociate ("playlist position", ui_playlist_notebook_position);
+    hook_dissociate ("playlist update", pl_notebook_update);
+    hook_dissociate ("playlist activate", pl_notebook_activate);
+    hook_dissociate ("playlist set playing", pl_notebook_set_playing);
+    hook_dissociate ("playlist position", pl_notebook_set_position);
+    hook_dissociate ("enable record", update_toggles);
+    hook_dissociate ("set record", update_toggles);
     hook_dissociate ("set shuffle", update_toggles);
     hook_dissociate ("set repeat", update_toggles);
-    hook_dissociate ("enable record", record_toggled);
     hook_dissociate ("config save", (HookFunction) config_save);
 }
 
@@ -804,16 +798,24 @@ bool GtkUI::init ()
         aud_plugin_add_watch (search_tool, search_tool_toggled, nullptr);
     }
 
-    /* playback buttons */
+    /* open/add buttons */
     toolbar_button_add (toolbar, button_open_pressed, "document-open");
     toolbar_button_add (toolbar, button_add_pressed, "list-add");
+
+    gtk_toolbar_insert ((GtkToolbar *) toolbar, gtk_separator_tool_item_new (), -1);
+
+    /* playback buttons */
     toolbar_button_add (toolbar, aud_drct_pl_prev, "media-skip-backward");
-    toolbar_button_add (toolbar, aud_drct_pl_next, "media-skip-forward");
     button_play = toolbar_button_add (toolbar, aud_drct_play_pause, "media-playback-start");
     button_stop = toolbar_button_add (toolbar, aud_drct_stop, "media-playback-stop");
+    toolbar_button_add (toolbar, aud_drct_pl_next, "media-skip-forward");
+
     button_record = toggle_button_new ("media-record", toggle_record);
+    gtk_widget_set_no_show_all ((GtkWidget *) button_record, true);
     gtk_toolbar_insert ((GtkToolbar *) toolbar, button_record, -1);
 
+    gtk_toolbar_insert ((GtkToolbar *) toolbar, gtk_separator_tool_item_new (), -1);
+
     /* time slider and label */
     GtkToolItem * boxitem1 = gtk_tool_item_new ();
     gtk_tool_item_set_expand (boxitem1, true);
@@ -837,6 +839,8 @@ bool GtkUI::init ()
     gtk_widget_set_no_show_all (slider, true);
     gtk_widget_set_no_show_all (label_time, true);
 
+    gtk_toolbar_insert ((GtkToolbar *) toolbar, gtk_separator_tool_item_new (), -1);
+
     /* repeat and shuffle buttons */
     button_repeat = toggle_button_new ("media-playlist-repeat", toggle_repeat);
     gtk_toolbar_insert ((GtkToolbar *) toolbar, button_repeat, -1);
@@ -870,8 +874,7 @@ bool GtkUI::init ()
     vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
     layout_add_center (vbox);
 
-    ui_playlist_notebook_new ();
-    gtk_box_pack_start ((GtkBox *) vbox, (GtkWidget *) UI_PLAYLIST_NOTEBOOK, true, true, 0);
+    gtk_box_pack_start ((GtkBox *) vbox, pl_notebook_new (), true, true, 0);
 
     /* optional UI elements */
     show_hide_menu ();
@@ -882,7 +885,7 @@ bool GtkUI::init ()
     ui_hooks_associate ();
 
     AUDDBG ("playlist associate\n");
-    ui_playlist_notebook_populate ();
+    pl_notebook_populate ();
 
     g_signal_connect (slider, "change-value", (GCallback) ui_slider_change_value_cb , nullptr);
     g_signal_connect (slider, "button-press-event", (GCallback) ui_slider_button_press_cb, nullptr);
@@ -894,10 +897,10 @@ bool GtkUI::init ()
 
     timer_add (TimerRate::Hz4, ui_volume_slider_update, volume);
 
-    g_signal_connect (window, "map-event", (GCallback) window_mapped_cb, nullptr);
+    g_signal_connect (window, "map-event", (GCallback) pl_notebook_grab_focus, nullptr);
     g_signal_connect (window, "delete-event", (GCallback) window_delete, nullptr);
     g_signal_connect (window, "key-press-event", (GCallback) window_keypress_cb, nullptr);
-    g_signal_connect (UI_PLAYLIST_NOTEBOOK, "key-press-event", (GCallback) playlist_keypress_cb, nullptr);
+    g_signal_connect (pl_notebook, "key-press-event", (GCallback) playlist_keypress_cb, nullptr);
 
     if (aud_drct_get_playing ())
     {
@@ -911,7 +914,6 @@ bool GtkUI::init ()
     title_change ();
 
     update_toggles ();
-    record_toggled ();
 
     gtk_widget_show_all (vbox_outer);
 
@@ -1073,9 +1075,9 @@ void popup_menu_rclick (unsigned button, uint32_t time)
      time);
 }
 
-void popup_menu_tab (unsigned button, uint32_t time, int playlist)
+void popup_menu_tab (unsigned button, uint32_t time, Playlist playlist)
 {
-    menu_tab_playlist_id = aud_playlist_get_unique_id (playlist);
+    menu_tab_playlist = playlist;
     gtk_menu_popup ((GtkMenu *) menu_tab, nullptr, nullptr, nullptr, nullptr, button, time);
 }
 
