@@ -1,6 +1,6 @@
 /*
  * ui_playlist_notebook.c
- * Copyright 2010-2012 Michał Lipski and John Lindgren
+ * Copyright 2010-2017 Michał Lipski and John Lindgren
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -22,6 +22,7 @@
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
+#define AUD_GLIB_INTEGRATION
 #include <libaudcore/runtime.h>
 #include <libaudcore/playlist.h>
 #include <libaudcore/audstrings.h>
@@ -29,17 +30,28 @@
 #include <libaudgui/list.h>
 #include <libaudgui/libaudgui.h>
 
+#include "../ui-common/menu-ops.h"
+
 #include "gtkui.h"
 #include "ui_playlist_notebook.h"
 #include "ui_playlist_widget.h"
-#include "playlist_util.h"
 
-static GtkWidget * notebook = nullptr;
-static int highlighted = -1;
+GtkWidget * pl_notebook = nullptr;
+
+static Playlist highlighted;
 
 static int switch_handler = 0;
 static int reorder_handler = 0;
 
+static GtkWidget * treeview_of (GtkWidget * page)
+    { return (GtkWidget *) g_object_get_data ((GObject *) page, "treeview"); }
+
+static GtkWidget * treeview_at_idx (int idx)
+    { return treeview_of (gtk_notebook_get_nth_page ((GtkNotebook *) pl_notebook, idx)); }
+
+static Playlist list_of (GtkWidget * widget)
+    { return aud::from_ptr<Playlist> (g_object_get_data ((GObject *) widget, "playlist")); }
+
 void apply_column_widths (GtkWidget * treeview)
 {
     /* skip righthand column since it expands with the window */
@@ -53,9 +65,9 @@ void apply_column_widths (GtkWidget * treeview)
 
 static void size_allocate_cb (GtkWidget * treeview)
 {
-    int current = gtk_notebook_get_current_page ((GtkNotebook *) notebook);
+    int current = gtk_notebook_get_current_page ((GtkNotebook *) pl_notebook);
 
-    if (current < 0 || treeview != playlist_get_treeview (current))
+    if (current < 0 || treeview != treeview_at_idx (current))
         return;
 
     bool changed = false;
@@ -75,12 +87,12 @@ static void size_allocate_cb (GtkWidget * treeview)
 
     if (changed)
     {
-        int count = gtk_notebook_get_n_pages ((GtkNotebook *) notebook);
+        int count = gtk_notebook_get_n_pages ((GtkNotebook *) pl_notebook);
 
         for (int i = 0; i < count; i ++)
         {
             if (i != current)
-                apply_column_widths (playlist_get_treeview (i));
+                apply_column_widths (treeview_at_idx (i));
         }
     }
 }
@@ -93,54 +105,36 @@ static void make_add_button (GtkWidget * notebook)
      ("list-add", GTK_ICON_SIZE_MENU));
     gtk_widget_set_can_focus (button, false);
 
-    g_signal_connect (button, "clicked", (GCallback) aud_playlist_new, nullptr);
+    g_signal_connect (button, "clicked", pl_new, nullptr);
     gtk_widget_show_all (button);
 
     gtk_notebook_set_action_widget ((GtkNotebook *) notebook, button, GTK_PACK_END);
 }
 
-static void close_button_cb (GtkWidget * button, void * id)
+static void close_button_cb (GtkWidget * button, void * data)
 {
-    audgui_confirm_playlist_delete (aud_playlist_by_unique_id (GPOINTER_TO_INT (id)));
+    audgui_confirm_playlist_delete (aud::from_ptr<Playlist> (data));
 }
 
-static GtkWidget * make_close_button (GtkWidget * ebox, int list)
+static GtkWidget * make_close_button (GtkWidget * ebox, Playlist list)
 {
     GtkWidget * button = gtk_button_new ();
     GtkWidget * image = gtk_image_new_from_icon_name ("window-close", GTK_ICON_SIZE_MENU);
     gtk_button_set_image ((GtkButton *) button, image);
     gtk_button_set_relief ((GtkButton *) button, GTK_RELIEF_NONE);
-    gtk_button_set_focus_on_click ((GtkButton *) button, false);
-    gtk_widget_set_name (button, "gtkui-tab-close-button");
+    gtk_widget_set_can_focus (button, false);
 
-    g_signal_connect (button, "clicked", (GCallback) close_button_cb,
-     GINT_TO_POINTER (aud_playlist_get_unique_id (list)));
-
-    GtkCssProvider * provider = gtk_css_provider_new ();
-    gtk_css_provider_load_from_data (provider,
-     "#gtkui-tab-close-button {"
-     " -GtkButton-default-border: 0;"
-     " -GtkButton-default-outside-border: 0;"
-     " -GtkButton-inner-border: 0;"
-     " -GtkWidget-focus-padding: 0;"
-     " -GtkWidget-focus-line-width: 0;"
-     " margin: 0;"
-     " padding: 0; }",
-     -1, nullptr);
-
-    gtk_style_context_add_provider (gtk_widget_get_style_context (button),
-     GTK_STYLE_PROVIDER (provider),
-     GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
-    g_object_unref (provider);
+    g_signal_connect (button, "clicked", (GCallback) close_button_cb, aud::to_ptr (list));
 
     gtk_widget_show (button);
 
     return button;
 }
 
-GtkNotebook * ui_playlist_get_notebook ()
+void pl_notebook_grab_focus ()
 {
-    return (GtkNotebook *) notebook;
+    int idx = gtk_notebook_get_current_page ((GtkNotebook *) pl_notebook);
+    gtk_widget_grab_focus (treeview_at_idx (idx));
 }
 
 static void tab_title_reset (GtkWidget * ebox)
@@ -151,12 +145,10 @@ static void tab_title_reset (GtkWidget * ebox)
     gtk_widget_show (label);
 }
 
-static void tab_title_save (GtkEntry * entry, void * ebox)
+static void tab_title_save (GtkEntry * entry, GtkWidget * ebox)
 {
-    int id = GPOINTER_TO_INT (g_object_get_data ((GObject *) ebox, "playlist-id"));
     GtkWidget * label = (GtkWidget *) g_object_get_data ((GObject *) ebox, "label");
-
-    aud_playlist_set_title (aud_playlist_by_unique_id (id), gtk_entry_get_text (entry));
+    list_of (ebox).set_title (gtk_entry_get_text (entry));
     gtk_widget_hide ((GtkWidget *) entry);
     gtk_widget_show (label);
 }
@@ -171,17 +163,16 @@ static gboolean tab_key_press_cb (GtkWidget * widget, GdkEventKey * event)
 
 static gboolean tab_button_press_cb (GtkWidget * ebox, GdkEventButton * event)
 {
-    int id = GPOINTER_TO_INT (g_object_get_data ((GObject *) ebox, "playlist-id"));
-    int playlist = aud_playlist_by_unique_id (id);
+    auto list = list_of (ebox);
 
     if (event->type == GDK_2BUTTON_PRESS && event->button == 1)
-        aud_playlist_play (playlist);
+        list.start_playback ();
 
     if (event->type == GDK_BUTTON_PRESS && event->button == 2)
-        audgui_confirm_playlist_delete (playlist);
+        audgui_confirm_playlist_delete (list);
 
     if (event->type == GDK_BUTTON_PRESS && event->button == 3)
-        popup_menu_tab (event->button, event->time, playlist);
+        popup_menu_tab (event->button, event->time, list);
 
     return false;
 }
@@ -192,12 +183,12 @@ static gboolean scroll_cb (GtkWidget * widget, GdkEventScroll * event)
     {
     case GDK_SCROLL_UP:
     case GDK_SCROLL_LEFT:
-        aud_playlist_set_active (aud_playlist_get_active () - 1);
+        pl_prev ();
         return true;
 
     case GDK_SCROLL_DOWN:
     case GDK_SCROLL_RIGHT:
-        aud_playlist_set_active (aud_playlist_get_active () + 1);
+        pl_next ();
         return true;
 
     default:
@@ -212,63 +203,61 @@ static gboolean scroll_ignore_cb ()
 
 static void tab_changed (GtkNotebook * notebook, GtkWidget * page, unsigned page_num)
 {
-    aud_playlist_set_active (page_num);
+    Playlist::by_index (page_num).activate ();
 }
 
-static void tab_reordered (GtkNotebook * notebook, GtkWidget * child, unsigned page_num)
+static void tab_reordered (GtkNotebook * notebook, GtkWidget * page, unsigned page_num)
 {
-    GtkWidget * widget = (GtkWidget *) g_object_get_data ((GObject *) child, "treeview");
-    g_return_if_fail (widget);
-    aud_playlist_reorder (ui_playlist_widget_get_playlist (widget), page_num, 1);
+    auto list = list_of (treeview_of (page));
+    Playlist::reorder_playlists (list.index (), page_num, 1);
 }
 
-static GtkLabel * get_tab_label (int playlist)
+static GtkLabel * get_tab_label (int list_idx)
 {
-    GtkWidget * page = gtk_notebook_get_nth_page (UI_PLAYLIST_NOTEBOOK, playlist);
-    GtkWidget * ebox = gtk_notebook_get_tab_label (UI_PLAYLIST_NOTEBOOK, page);
+    GtkWidget * page = gtk_notebook_get_nth_page ((GtkNotebook *) pl_notebook, list_idx);
+    GtkWidget * ebox = gtk_notebook_get_tab_label ((GtkNotebook *) pl_notebook, page);
     return (GtkLabel *) g_object_get_data ((GObject *) ebox, "label");
 }
 
-static void set_tab_label (int list, GtkLabel * label)
+static void update_tab_label (GtkLabel * label, Playlist list)
 {
-    String title0 = aud_playlist_get_title (list);
+    String title0 = list.get_title ();
     StringBuf title = aud_get_bool ("gtkui", "entry_count_visible") ?
-     str_printf ("%s (%d)", (const char *) title0, aud_playlist_entry_count (list)) :
+     str_printf ("%s (%d)", (const char *) title0, list.n_entries ()) :
      str_copy (title0);
 
-    if (list == aud_playlist_get_playing ())
+    if (list == Playlist::playing_playlist ())
     {
-        char * markup = g_markup_printf_escaped ("<b>%s</b>", (const char *) title);
+        CharPtr markup (g_markup_printf_escaped ("<b>%s</b>", (const char *) title));
         gtk_label_set_markup (label, markup);
-        g_free (markup);
     }
     else
         gtk_label_set_text (label, title);
 }
 
-void start_rename_playlist (int playlist)
+void start_rename_playlist (Playlist playlist)
 {
-    if (! gtk_notebook_get_show_tabs ((GtkNotebook *) notebook))
+    if (! gtk_notebook_get_show_tabs ((GtkNotebook *) pl_notebook))
     {
         audgui_show_playlist_rename (playlist);
         return;
     }
 
-    GtkWidget * page = gtk_notebook_get_nth_page (UI_PLAYLIST_NOTEBOOK, playlist);
-    GtkWidget * ebox = gtk_notebook_get_tab_label (UI_PLAYLIST_NOTEBOOK, page);
+    GtkWidget * page = gtk_notebook_get_nth_page ((GtkNotebook *) pl_notebook, playlist.index ());
+    GtkWidget * ebox = gtk_notebook_get_tab_label ((GtkNotebook *) pl_notebook, page);
 
     GtkWidget * label = (GtkWidget *) g_object_get_data ((GObject *) ebox, "label");
     GtkWidget * entry = (GtkWidget *) g_object_get_data ((GObject *) ebox, "entry");
     gtk_widget_hide (label);
 
-    gtk_entry_set_text ((GtkEntry *) entry, aud_playlist_get_title (playlist));
+    gtk_entry_set_text ((GtkEntry *) entry, playlist.get_title ());
 
     gtk_widget_grab_focus (entry);
     gtk_editable_select_region ((GtkEditable *) entry, 0, -1);
     gtk_widget_show (entry);
 }
 
-void ui_playlist_notebook_create_tab (int playlist)
+static void create_tab (int list_idx, Playlist list)
 {
     GtkWidget * scrollwin = gtk_scrolled_window_new (nullptr, nullptr);
     GtkAdjustment * vscroll = gtk_scrolled_window_get_vadjustment ((GtkScrolledWindow *) scrollwin);
@@ -276,7 +265,7 @@ void ui_playlist_notebook_create_tab (int playlist)
     /* do not allow scroll events to propagate up to the notebook */
     g_signal_connect_after (scrollwin, "scroll-event", (GCallback) scroll_ignore_cb, nullptr);
 
-    GtkWidget * treeview = ui_playlist_widget_new (playlist);
+    GtkWidget * treeview = ui_playlist_widget_new (list);
 
     apply_column_widths (treeview);
     g_signal_connect (treeview, "size-allocate", (GCallback) size_allocate_cb, nullptr);
@@ -294,7 +283,7 @@ void ui_playlist_notebook_create_tab (int playlist)
     GtkWidget * hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 2);
 
     GtkWidget * label = gtk_label_new ("");
-    set_tab_label (playlist, (GtkLabel *) label);
+    update_tab_label ((GtkLabel *) label, list);
     gtk_box_pack_start ((GtkBox *) hbox, label, false, false, 0);
 
     GtkWidget * entry = gtk_entry_new ();
@@ -307,7 +296,7 @@ void ui_playlist_notebook_create_tab (int playlist)
 
     if (aud_get_bool ("gtkui", "close_button_visible"))
     {
-        button = make_close_button (ebox, playlist);
+        button = make_close_button (ebox, list);
         gtk_box_pack_end ((GtkBox *) hbox, button, false, false, 0);
     }
 
@@ -315,18 +304,17 @@ void ui_playlist_notebook_create_tab (int playlist)
     g_object_set_data ((GObject *) ebox, "entry", entry);
     g_object_set_data ((GObject *) ebox, "page", scrollwin);
 
-    gtk_notebook_insert_page (UI_PLAYLIST_NOTEBOOK, scrollwin, ebox, playlist);
-    gtk_notebook_set_tab_reorderable (UI_PLAYLIST_NOTEBOOK, scrollwin, true);
+    gtk_notebook_insert_page ((GtkNotebook *) pl_notebook, scrollwin, ebox, list_idx);
+    gtk_notebook_set_tab_reorderable ((GtkNotebook *) pl_notebook, scrollwin, true);
 
-    int id = aud_playlist_get_unique_id (playlist);
-    g_object_set_data ((GObject *) ebox, "playlist-id", GINT_TO_POINTER (id));
-    g_object_set_data ((GObject *) treeview, "playlist-id", GINT_TO_POINTER (id));
+    g_object_set_data ((GObject *) ebox, "playlist", aud::to_ptr (list));
+    g_object_set_data ((GObject *) treeview, "playlist", aud::to_ptr (list));
 
-    int position = aud_playlist_get_position (playlist);
+    int position = list.get_position ();
     if (position >= 0)
         audgui_list_set_highlight (treeview, position);
 
-    int focus = aud_playlist_get_focus (playlist);
+    int focus = list.get_focus ();
     if (focus >= 0)
         audgui_list_set_focus (treeview, position);
 
@@ -348,70 +336,71 @@ void ui_playlist_notebook_create_tab (int playlist)
     }
 }
 
-void ui_playlist_notebook_populate ()
+static void switch_to_active ()
 {
-    int playlist = aud_playlist_get_active ();
-    int playlist_count = aud_playlist_count ();
-
-    for (int count = 0; count < playlist_count; count ++)
-        ui_playlist_notebook_create_tab (count);
-
-    gtk_notebook_set_current_page (UI_PLAYLIST_NOTEBOOK, playlist);
-    highlighted = aud_playlist_get_unique_id (aud_playlist_get_playing ());
-
-    if (! switch_handler)
-        switch_handler = g_signal_connect (notebook, "switch-page", (GCallback)
-         tab_changed, nullptr);
-    if (! reorder_handler)
-        reorder_handler = g_signal_connect (notebook, "page-reordered",
-         (GCallback) tab_reordered, nullptr);
-
-    gtk_widget_grab_focus (playlist_get_treeview (playlist));
+    int active_idx = Playlist::active_playlist ().index ();
+    gtk_notebook_set_current_page ((GtkNotebook *) pl_notebook, active_idx);
 }
 
-void ui_playlist_notebook_empty ()
+void pl_notebook_populate ()
+{
+    int n_playlists = Playlist::n_playlists ();
+    for (int idx = 0; idx < n_playlists; idx ++)
+        create_tab (idx, Playlist::by_index (idx));
+
+    switch_to_active ();
+    highlighted = Playlist::playing_playlist ();
+
+    if (! switch_handler)
+        switch_handler = g_signal_connect (pl_notebook, "switch-page",
+         (GCallback) tab_changed, nullptr);
+    if (! reorder_handler)
+        reorder_handler = g_signal_connect (pl_notebook, "page-reordered",
+         (GCallback) tab_reordered, nullptr);
+
+    pl_notebook_grab_focus ();
+}
+
+void pl_notebook_purge ()
 {
     if (switch_handler)
-        g_signal_handler_disconnect (notebook, switch_handler);
+        g_signal_handler_disconnect (pl_notebook, switch_handler);
     switch_handler = 0;
     if (reorder_handler)
-        g_signal_handler_disconnect (notebook, reorder_handler);
+        g_signal_handler_disconnect (pl_notebook, reorder_handler);
     reorder_handler = 0;
 
-    int n_pages = gtk_notebook_get_n_pages ((GtkNotebook *) notebook);
+    int n_pages = gtk_notebook_get_n_pages ((GtkNotebook *) pl_notebook);
     while (n_pages)
-        gtk_notebook_remove_page ((GtkNotebook *) notebook, -- n_pages);
+        gtk_notebook_remove_page ((GtkNotebook *) pl_notebook, -- n_pages);
 }
 
 static void add_remove_pages ()
 {
-    g_signal_handlers_block_by_func (notebook, (void *) tab_changed, nullptr);
-    g_signal_handlers_block_by_func (notebook, (void *) tab_reordered, nullptr);
+    g_signal_handlers_block_by_func (pl_notebook, (void *) tab_changed, nullptr);
+    g_signal_handlers_block_by_func (pl_notebook, (void *) tab_reordered, nullptr);
 
-    int lists = aud_playlist_count ();
-    int pages = gtk_notebook_get_n_pages ((GtkNotebook *) notebook);
+    int lists = Playlist::n_playlists ();
+    int pages = gtk_notebook_get_n_pages ((GtkNotebook *) pl_notebook);
 
     /* scan through existing treeviews */
     for (int i = 0; i < pages; )
     {
-        GtkWidget * page = gtk_notebook_get_nth_page ((GtkNotebook *) notebook, i);
-        GtkWidget * tree = (GtkWidget *) g_object_get_data ((GObject *) page, "treeview");
-        int tree_id = GPOINTER_TO_INT (g_object_get_data ((GObject *) tree, "playlist-id"));
+        auto list0 = list_of (treeview_at_idx (i));
 
         /* do we have an orphaned treeview? */
-        if (aud_playlist_by_unique_id (tree_id) < 0)
+        if (! list0.exists ())
         {
-            gtk_notebook_remove_page ((GtkNotebook *) notebook, i);
+            gtk_notebook_remove_page ((GtkNotebook *) pl_notebook, i);
             pages --;
             continue;
         }
 
         /* do we have the right treeview? */
-        int list_id = aud_playlist_get_unique_id (i);
+        auto list = Playlist::by_index (i);
 
-        if (tree_id == list_id)
+        if (list0 == list)
         {
-            ui_playlist_widget_set_playlist (tree, i);
             i ++;
             continue;
         }
@@ -421,14 +410,13 @@ static void add_remove_pages ()
 
         for (int j = i + 1; j < pages; j ++)
         {
-            page = gtk_notebook_get_nth_page ((GtkNotebook *) notebook, j);
-            tree = (GtkWidget *) g_object_get_data ((GObject *) page, "treeview");
-            tree_id = GPOINTER_TO_INT (g_object_get_data ((GObject *) tree, "playlist-id"));
+            GtkWidget * page = gtk_notebook_get_nth_page ((GtkNotebook *) pl_notebook, j);
+            auto list2 = list_of (treeview_of (page));
 
             /* found it? move it to the right place */
-            if (tree_id == list_id)
+            if (list2 == list)
             {
-                gtk_notebook_reorder_child ((GtkNotebook *) notebook, page, i);
+                gtk_notebook_reorder_child ((GtkNotebook *) pl_notebook, page, i);
                 found = true;
                 break;
             }
@@ -437,7 +425,7 @@ static void add_remove_pages ()
         /* didn't find it? create it */
         if (! found)
         {
-            ui_playlist_notebook_create_tab (i);
+            create_tab (i, list);
             pages ++;
             continue;
         }
@@ -446,115 +434,105 @@ static void add_remove_pages ()
     /* create new treeviews */
     while (pages < lists)
     {
-        ui_playlist_notebook_create_tab (pages);
+        create_tab (pages, Playlist::by_index (pages));
         pages ++;
     }
 
-    gtk_notebook_set_current_page ((GtkNotebook *) notebook, aud_playlist_get_active ());
-
+    switch_to_active ();
     show_hide_playlist_tabs ();
 
-    g_signal_handlers_unblock_by_func (notebook, (void *) tab_changed, nullptr);
-    g_signal_handlers_unblock_by_func (notebook, (void *) tab_reordered, nullptr);
+    g_signal_handlers_unblock_by_func (pl_notebook, (void *) tab_changed, nullptr);
+    g_signal_handlers_unblock_by_func (pl_notebook, (void *) tab_reordered, nullptr);
 }
 
-void ui_playlist_notebook_update (void * data, void * user)
+void pl_notebook_update (void * data, void * user)
 {
     auto global_level = aud::from_ptr<Playlist::UpdateLevel> (data);
     if (global_level == Playlist::Structure)
         add_remove_pages ();
 
-    int lists = aud_playlist_count ();
+    int n_pages = gtk_notebook_get_n_pages ((GtkNotebook *) pl_notebook);
 
-    for (int list = 0; list < lists; list ++)
+    for (int i = 0; i < n_pages; i ++)
     {
+        GtkWidget * treeview = treeview_at_idx (i);
+
         if (global_level >= Playlist::Metadata)
-            set_tab_label (list, get_tab_label (list));
+            update_tab_label (get_tab_label (i), list_of (treeview));
 
-        GtkWidget * treeview = playlist_get_treeview (list);
-
-        auto update = aud_playlist_update_detail (list);
-        if (update.level)
-            ui_playlist_widget_update (treeview, update);
-
-        audgui_list_set_highlight (treeview, aud_playlist_get_position (list));
+        ui_playlist_widget_update (treeview);
     }
 
-    gtk_notebook_set_current_page ((GtkNotebook *) notebook, aud_playlist_get_active ());
+    switch_to_active ();
 }
 
-void ui_playlist_notebook_position (void * data, void * user)
+void pl_notebook_set_position (void * data, void * user)
 {
-    int list = aud::from_ptr<int> (data);
-    int row = aud_playlist_get_position (list);
+    auto list = aud::from_ptr<Playlist> (data);
+    int row = list.get_position ();
 
     if (aud_get_bool ("gtkui", "autoscroll"))
     {
-        aud_playlist_select_all (list, false);
-        aud_playlist_entry_set_selected (list, row, true);
-        aud_playlist_set_focus (list, row);
+        list.select_all (false);
+        list.select_entry (row, true);
+        list.set_focus (row);
     }
 
-    if (! aud_playlist_update_pending ())
-        audgui_list_set_highlight (playlist_get_treeview (list), row);
+    audgui_list_set_highlight (treeview_at_idx (list.index ()), row);
 }
 
-void ui_playlist_notebook_activate (void * data, void * user)
+void pl_notebook_activate (void * data, void * user)
 {
-    if (! aud_playlist_update_pending ())
-        gtk_notebook_set_current_page ((GtkNotebook *) notebook, aud_playlist_get_active ());
+    switch_to_active ();
 }
 
-void ui_playlist_notebook_set_playing (void * data, void * user)
+void pl_notebook_set_playing (void * data, void * user)
 {
-    int id = aud_playlist_get_unique_id (aud_playlist_get_playing ());
+    auto playing = Playlist::playing_playlist ();
 
     // if the previous playing playlist was deleted, ignore it
-    if (aud_playlist_by_unique_id (highlighted) < 0)
-        highlighted = -1;
+    if (! highlighted.exists ())
+        highlighted = Playlist ();
 
-    if (highlighted == id)
+    if (highlighted == playing)
         return;
 
-    int pages = gtk_notebook_get_n_pages ((GtkNotebook *) notebook);
+    int pages = gtk_notebook_get_n_pages ((GtkNotebook *) pl_notebook);
 
     for (int i = 0; i < pages; i ++)
     {
-        GtkWidget * page = gtk_notebook_get_nth_page ((GtkNotebook *) notebook, i);
-        GtkWidget * tree = (GtkWidget *) g_object_get_data ((GObject *) page, "treeview");
-        int tree_id = GPOINTER_TO_INT (g_object_get_data ((GObject *) tree, "playlist-id"));
-
-        if (tree_id == highlighted || tree_id == id)
-            set_tab_label (i, get_tab_label (i));
+        auto list = list_of (treeview_at_idx (i));
+        if (list == highlighted || list == playing)
+            update_tab_label (get_tab_label (i), list);
     }
 
-    highlighted = id;
+    highlighted = playing;
 }
 
 static void destroy_cb ()
 {
-    notebook = nullptr;
+    pl_notebook = nullptr;
     switch_handler = 0;
     reorder_handler = 0;
 }
 
-GtkWidget * ui_playlist_notebook_new ()
+GtkWidget * pl_notebook_new ()
 {
-    notebook = gtk_notebook_new ();
-    gtk_notebook_set_scrollable ((GtkNotebook *) notebook, true);
-    make_add_button (notebook);
+    pl_notebook = gtk_notebook_new ();
+    gtk_notebook_set_scrollable ((GtkNotebook *) pl_notebook, true);
+    make_add_button (pl_notebook);
 
     show_hide_playlist_tabs ();
 
-    gtk_widget_add_events (notebook, GDK_SCROLL_MASK);
-    g_signal_connect (notebook, "scroll-event", (GCallback) scroll_cb, nullptr);
-    g_signal_connect (notebook, "destroy", (GCallback) destroy_cb, nullptr);
+    gtk_widget_add_events (pl_notebook, GDK_SCROLL_MASK);
+    g_signal_connect (pl_notebook, "scroll-event", (GCallback) scroll_cb, nullptr);
+    g_signal_connect (pl_notebook, "destroy", (GCallback) destroy_cb, nullptr);
 
-    return notebook;
+    return pl_notebook;
 }
 
 void show_hide_playlist_tabs ()
 {
-    gtk_notebook_set_show_tabs ((GtkNotebook *) notebook, aud_get_bool ("gtkui",
-     "playlist_tabs_visible") || aud_playlist_count () > 1);
+    gtk_notebook_set_show_tabs ((GtkNotebook *) pl_notebook, aud_get_bool ("gtkui",
+     "playlist_tabs_visible") || Playlist::n_playlists () > 1);
 }
