@@ -20,7 +20,7 @@
 #include <string.h>
 #include <gtk/gtk.h>
 
-#define AUD_PLUGIN_GLIB_ONLY
+#define AUD_GLIB_INTEGRATION
 #include <libaudcore/audstrings.h>
 #include <libaudcore/hook.h>
 #include <libaudcore/i18n.h>
@@ -41,7 +41,10 @@ class SearchTool : public GeneralPlugin
 public:
     static constexpr PluginInfo info = {
         N_("Search Tool"),
-        PACKAGE
+        PACKAGE,
+        nullptr, // about
+        nullptr, // prefs
+        PluginGLibOnly
     };
 
     constexpr SearchTool () : GeneralPlugin (info, false) {}
@@ -89,105 +92,99 @@ struct Item
     Item & operator= (Item &&) = default;
 };
 
-struct SearchState {
-    Index<const Item *> items;
-    int mask;
-};
-
-static int playlist_id;
-static Index<String> search_terms;
+static Playlist s_playlist;
+static Index<String> s_search_terms;
 
 /* Note: added_table is accessed by multiple threads.
  * When adding = true, it may only be accessed by the playlist add thread.
  * When adding = false, it may only be accessed by the UI thread.
  * adding may only be set by the UI thread while holding adding_lock. */
-static TinyLock adding_lock;
-static bool adding = false;
-static SimpleHash<String, bool> added_table;
+static TinyLock s_adding_lock;
+static bool s_adding = false;
+static SimpleHash<String, bool> s_added_table;
 
-static SimpleHash<Key, Item> database;
-static bool database_valid;
-static Index<const Item *> items;
-static int hidden_items;
-static Index<bool> selection;
+static SimpleHash<Key, Item> s_database;
+static bool s_database_valid;
+static Index<const Item *> s_items;
+static int s_hidden_items;
+static Index<bool> s_selection;
 
-static QueuedFunc search_timer;
-static bool search_pending;
+static QueuedFunc s_search_timer;
+static bool s_search_pending;
 
 static GtkWidget * entry, * help_label, * wait_label, * scrolled, * results_list, * stats_label;
 
 static void find_playlist ()
 {
-    playlist_id = -1;
+    s_playlist = Playlist ();
 
-    for (int p = 0; playlist_id < 0 && p < aud_playlist_count (); p ++)
+    for (int p = 0; p < Playlist::n_playlists (); p ++)
     {
-        String title = aud_playlist_get_title (p);
-        if (! strcmp (title, _("Library")))
-            playlist_id = aud_playlist_get_unique_id (p);
+        auto playlist = Playlist::by_index (p);
+        if (! strcmp (playlist.get_title (), _("Library")))
+        {
+            s_playlist = playlist;
+            break;
+        }
     }
 }
 
-static int create_playlist ()
+static void create_playlist ()
 {
-    int list = aud_playlist_get_blank ();
-    aud_playlist_set_title (list, _("Library"));
-    aud_playlist_set_active (list);
-    playlist_id = aud_playlist_get_unique_id (list);
-    return list;
+    s_playlist = Playlist::blank_playlist ();
+    s_playlist.set_title (_("Library"));
+    s_playlist.active_playlist ();
 }
 
-static int get_playlist (bool require_added, bool require_scanned)
+static bool check_playlist (bool require_added, bool require_scanned)
 {
-    if (playlist_id < 0)
-        return -1;
-
-    int list = aud_playlist_by_unique_id (playlist_id);
-
-    if (list < 0)
+    if (! s_playlist.exists ())
     {
-        playlist_id = -1;
-        return -1;
+        s_playlist = Playlist ();
+        return false;
     }
 
-    if (require_added && aud_playlist_add_in_progress (list))
-        return -1;
-    if (require_scanned && aud_playlist_scan_in_progress (list))
-        return -1;
+    if (require_added && s_playlist.add_in_progress ())
+        return false;
+    if (require_scanned && s_playlist.scan_in_progress ())
+        return false;
 
-    return list;
+    return true;
 }
 
-static String get_path ()
+static String get_uri ()
 {
+    auto to_uri = [] (const char * path)
+        { return String (filename_to_uri (path)); };
+
     String path1 = aud_get_str ("search-tool", "path");
-    if (g_file_test (path1, G_FILE_TEST_EXISTS))
-        return path1;
+    if (path1[0])
+        return strstr (path1, "://") ? path1 : to_uri (path1);
 
     StringBuf path2 = filename_build ({g_get_home_dir (), "Music"});
     if (g_file_test (path2, G_FILE_TEST_EXISTS))
-        return String (path2);
+        return to_uri (path2);
 
-    return String (g_get_home_dir ());
+    return to_uri (g_get_home_dir ());
 }
 
 static void destroy_database ()
 {
-    items.clear ();
-    hidden_items = 0;
-    database.clear ();
-    database_valid = false;
+    s_items.clear ();
+    s_hidden_items = 0;
+    s_database.clear ();
+    s_database_valid = false;
 }
 
-static void create_database (int list)
+static void create_database ()
 {
     destroy_database ();
 
-    int entries = aud_playlist_entry_count (list);
+    int entries = s_playlist.n_entries ();
 
     for (int e = 0; e < entries; e ++)
     {
-        Tuple tuple = aud_playlist_entry_get_tuple (list, e, Playlist::Nothing);
+        Tuple tuple = s_playlist.entry_tuple (e, Playlist::NoWait);
 
         aud::array<SearchField, String> fields;
         fields[SearchField::Genre] = tuple.get_str (Tuple::Genre);
@@ -196,7 +193,7 @@ static void create_database (int list)
         fields[SearchField::Title] = tuple.get_str (Tuple::Title);
 
         Item * parent = nullptr;
-        SimpleHash<Key, Item> * hash = & database;
+        SimpleHash<Key, Item> * hash = & s_database;
 
         for (auto f : aud::range<SearchField> ())
         {
@@ -220,37 +217,36 @@ static void create_database (int list)
         }
     }
 
-    database_valid = true;
+    s_database_valid = true;
 }
 
-static void search_cb (const Key & key, Item & item, void * _state)
+static void search_recurse (SimpleHash<Key, Item> & domain, int mask, Index<const Item *> & results)
 {
-    SearchState * state = (SearchState *) _state;
-
-    int oldmask = state->mask;
-    int count = search_terms.len ();
-
-    for (int t = 0, bit = 1; t < count; t ++, bit <<= 1)
+    domain.iterate ([mask, & results] (const Key & key, Item & item)
     {
-        if (! (state->mask & bit))
-            continue; /* skip term if it is already found */
+        int count = s_search_terms.len ();
+        int new_mask = mask;
 
-        if (strstr (item.folded, search_terms[t]))
-            state->mask &= ~bit; /* we found it */
-        else if (! item.children.n_items ())
-            break; /* quit early if there are no children to search */
-    }
+        for (int t = 0, bit = 1; t < count; t ++, bit <<= 1)
+        {
+            if (! (new_mask & bit))
+                continue; /* skip term if it is already found */
 
-    /* adding an item with exactly one child is redundant, so avoid it */
-    if (! state->mask && item.children.n_items () != 1)
-        state->items.append (& item);
+            if (strstr (item.folded, s_search_terms[t]))
+                new_mask &= ~bit; /* we found it */
+            else if (! item.children.n_items ())
+                break; /* quit early if there are no children to search */
+        }
 
-    item.children.iterate (search_cb, state);
+        /* adding an item with exactly one child is redundant, so avoid it */
+        if (! new_mask && item.children.n_items () != 1)
+            results.append (& item);
 
-    state->mask = oldmask;
+        search_recurse (item.children, new_mask, results);
+    });
 }
 
-static int item_compare (const Item * const & a, const Item * const & b, void *)
+static int item_compare (const Item * const & a, const Item * const & b)
 {
     if (a->field < b->field)
         return -1;
@@ -262,115 +258,113 @@ static int item_compare (const Item * const & a, const Item * const & b, void *)
         return val;
 
     if (a->parent)
-        return b->parent ? item_compare (a->parent, b->parent, nullptr) : 1;
+        return b->parent ? item_compare (a->parent, b->parent) : 1;
     else
         return b->parent ? -1 : 0;
 }
 
-static int item_compare_pass1 (const Item * const & a, const Item * const & b, void *)
+static int item_compare_pass1 (const Item * const & a, const Item * const & b)
 {
     if (a->matches.len () > b->matches.len ())
         return -1;
     if (a->matches.len () < b->matches.len ())
         return 1;
 
-    return item_compare (a, b, nullptr);
+    return item_compare (a, b);
 }
 
 static void do_search ()
 {
-    items.clear ();
-    hidden_items = 0;
+    s_items.clear ();
+    s_hidden_items = 0;
 
-    if (! database_valid)
+    if (! s_database_valid)
         return;
 
-    SearchState state;
-
     /* effectively limits number of search terms to 32 */
-    state.mask = (1 << search_terms.len ()) - 1;
-
-    database.iterate (search_cb, & state);
-
-    items = std::move (state.items);
+    search_recurse (s_database, (1 << s_search_terms.len ()) - 1, s_items);
 
     /* first sort by number of songs per item */
-    items.sort (item_compare_pass1, nullptr);
+    s_items.sort (item_compare_pass1);
 
     /* limit to items with most songs */
-    if (items.len () > MAX_RESULTS)
+    if (s_items.len () > MAX_RESULTS)
     {
-        hidden_items = items.len () - MAX_RESULTS;
-        items.remove (MAX_RESULTS, -1);
+        s_hidden_items = s_items.len () - MAX_RESULTS;
+        s_items.remove (MAX_RESULTS, -1);
     }
 
     /* sort by item type, then item name */
-    items.sort (item_compare, nullptr);
+    s_items.sort (item_compare);
 
-    selection.remove (0, -1);
-    selection.insert (0, items.len ());
-    if (items.len ())
-        selection[0] = true;
+    s_selection.remove (0, -1);
+    s_selection.insert (0, s_items.len ());
+    if (s_items.len ())
+        s_selection[0] = true;
 }
 
 static bool filter_cb (const char * filename, void * unused)
 {
-    tiny_lock (& adding_lock);
-    bool add = adding && ! added_table.lookup (String (filename));
-    tiny_unlock (& adding_lock);
+    bool add = false;
+    tiny_lock (& s_adding_lock);
+
+    if (s_adding)
+    {
+        bool * added = s_added_table.lookup (String (filename));
+
+        if ((add = ! added))
+            s_added_table.add (String (filename), true);
+        else
+            (* added) = true;
+    }
+
+    tiny_unlock (& s_adding_lock);
     return add;
 }
 
-static void begin_add (const char * path)
+static void begin_add (const char * uri)
 {
-    if (adding)
+    if (s_adding)
         return;
 
-    int list = get_playlist (false, false);
+    if (! check_playlist (false, false))
+        create_playlist ();
 
-    if (list < 0)
-        list = create_playlist ();
+    /* if possible, store local path for compatibility with older versions */
+    StringBuf path = uri_to_filename (uri);
+    aud_set_str ("search-tool", "path", path ? path : uri);
 
-    aud_set_str ("search-tool", "path", path);
+    s_added_table.clear ();
 
-    StringBuf uri = filename_to_uri (path);
-    g_return_if_fail (uri);
-
-    if (! g_str_has_suffix (uri, "/"))
-        uri.insert (-1, "/");
-
-    added_table.clear ();
-
-    int entries = aud_playlist_entry_count (list);
+    int entries = s_playlist.n_entries ();
 
     for (int entry = 0; entry < entries; entry ++)
     {
-        String filename = aud_playlist_entry_get_filename (list, entry);
+        String filename = s_playlist.entry_filename (entry);
 
-        if (g_str_has_prefix (filename, uri) && ! added_table.lookup (filename))
+        if (! s_added_table.lookup (filename))
         {
-            aud_playlist_entry_set_selected (list, entry, false);
-            added_table.add (filename, true);
+            s_playlist.select_entry (entry, false);
+            s_added_table.add (filename, false);
         }
         else
-            aud_playlist_entry_set_selected (list, entry, true);
+            s_playlist.select_entry (entry, true);
     }
 
-    aud_playlist_delete_selected (list);
-    aud_playlist_remove_failed (list);
+    s_playlist.remove_selected ();
 
-    tiny_lock (& adding_lock);
-    adding = true;
-    tiny_unlock (& adding_lock);
+    tiny_lock (& s_adding_lock);
+    s_adding = true;
+    tiny_unlock (& s_adding_lock);
 
     Index<PlaylistAddItem> add;
     add.append (String (uri));
-    aud_playlist_entry_insert_filtered (list, -1, std::move (add), filter_cb, nullptr, false);
+    s_playlist.insert_filtered (-1, std::move (add), filter_cb, nullptr, false);
 }
 
 static void show_hide_widgets ()
 {
-    if (playlist_id < 0)
+    if (s_playlist == Playlist ())
     {
         gtk_widget_hide (wait_label);
         gtk_widget_hide (scrolled);
@@ -381,7 +375,7 @@ static void show_hide_widgets ()
     {
         gtk_widget_hide (help_label);
 
-        if (database_valid)
+        if (s_database_valid)
         {
             gtk_widget_hide (wait_label);
             gtk_widget_show (scrolled);
@@ -401,32 +395,27 @@ static void search_timeout (void * = nullptr)
     do_search ();
 
     audgui_list_delete_rows (results_list, 0, audgui_list_row_count (results_list));
-    audgui_list_insert_rows (results_list, 0, items.len ());
+    audgui_list_insert_rows (results_list, 0, s_items.len ());
 
-    int total = items.len () + hidden_items;
-    StringBuf stats = str_printf (dngettext (PACKAGE, "%d result",
-     "%d results", total), total);
+    int total = s_items.len () + s_hidden_items;
 
-    if (hidden_items)
-    {
-        stats.insert (-1, " ");
-        stats.combine (str_printf (dngettext (PACKAGE, "(%d hidden)",
-         "(%d hidden)", hidden_items), hidden_items));
-    }
+    if (s_hidden_items)
+        gtk_label_set_text ((GtkLabel *) stats_label,
+         str_printf (dngettext (PACKAGE, "%d of %d result shown",
+         "%d of %d results shown", total), s_items.len (), total));
+    else
+        gtk_label_set_text ((GtkLabel *) stats_label,
+         str_printf (dngettext (PACKAGE, "%d result", "%d results", total), total));
 
-    gtk_label_set_text ((GtkLabel *) stats_label, stats);
-
-    search_timer.stop ();
-    search_pending = false;
+    s_search_timer.stop ();
+    s_search_pending = false;
 }
 
 static void update_database ()
 {
-    int list = get_playlist (true, true);
-
-    if (list >= 0)
+    if (check_playlist (true, true))
     {
-        create_database (list);
+        create_database ();
         search_timeout ();
     }
     else
@@ -439,45 +428,57 @@ static void update_database ()
     show_hide_widgets ();
 }
 
-static void add_complete_cb (void * unused, void * unused2)
+static void add_complete_cb (void *, void *)
 {
-    int list = get_playlist (true, false);
-    if (list < 0)
+    if (! check_playlist (true, false))
         return;
 
-    if (adding)
+    if (s_adding)
     {
-        tiny_lock (& adding_lock);
-        adding = false;
-        tiny_unlock (& adding_lock);
+        tiny_lock (& s_adding_lock);
+        s_adding = false;
+        tiny_unlock (& s_adding_lock);
 
-        added_table.clear ();
-        aud_playlist_sort_by_scheme (list, Playlist::Path);
+        int entries = s_playlist.n_entries ();
+
+        for (int entry = 0; entry < entries; entry ++)
+        {
+            String filename = s_playlist.entry_filename (entry);
+            bool * added = s_added_table.lookup (filename);
+
+            s_playlist.select_entry (entry, ! added || ! (* added));
+        }
+
+        s_added_table.clear ();
+
+        /* don't clear the playlist if nothing was added */
+        if (s_playlist.n_selected () < entries)
+            s_playlist.remove_selected ();
+        else
+            s_playlist.select_all (false);
+
+        s_playlist.sort_entries (Playlist::Path);
     }
 
-    if (! database_valid && ! aud_playlist_update_pending (list))
+    if (! s_database_valid && ! s_playlist.update_pending ())
         update_database ();
 }
 
-static void scan_complete_cb (void * unused, void * unused2)
+static void scan_complete_cb (void *, void *)
 {
-    int list = get_playlist (true, true);
-    if (list < 0)
+    if (! check_playlist (true, true))
         return;
 
-    if (! database_valid && ! aud_playlist_update_pending (list))
+    if (! s_database_valid && ! s_playlist.update_pending ())
         update_database ();
 }
 
-static void playlist_update_cb (void * data, void * unused)
+static void playlist_update_cb (void *, void *)
 {
-    if (! database_valid)
-        update_database ();
-    else
+    if (! s_database_valid || ! check_playlist (true, true) ||
+        s_playlist.update_detail ().level >= Playlist::Metadata)
     {
-        int list = get_playlist (true, true);
-        if (list < 0 || aud_playlist_update_detail (list).level >= Playlist::Metadata)
-            update_database ();
+        update_database ();
     }
 }
 
@@ -498,28 +499,27 @@ static void search_cleanup ()
     hook_dissociate ("playlist scan complete", scan_complete_cb);
     hook_dissociate ("playlist update", playlist_update_cb);
 
-    search_timer.stop ();
-    search_pending = false;
+    s_search_timer.stop ();
+    s_search_pending = false;
 
-    search_terms.clear ();
-    items.clear ();
-    selection.clear ();
+    s_search_terms.clear ();
+    s_items.clear ();
+    s_selection.clear ();
 
-    tiny_lock (& adding_lock);
-    adding = false;
-    tiny_unlock (& adding_lock);
+    tiny_lock (& s_adding_lock);
+    s_adding = false;
+    tiny_unlock (& s_adding_lock);
 
-    added_table.clear ();
+    s_added_table.clear ();
     destroy_database ();
 }
 
 static void do_add (bool play, bool set_title)
 {
-    if (search_pending)
+    if (s_search_pending)
         search_timeout ();
 
-    int list = aud_playlist_by_unique_id (playlist_id);
-    int n_items = items.len ();
+    int n_items = s_items.len ();
     int n_selected = 0;
 
     Index<PlaylistAddItem> add;
@@ -527,16 +527,17 @@ static void do_add (bool play, bool set_title)
 
     for (int i = 0; i < n_items; i ++)
     {
-        if (! selection[i])
+        if (! s_selection[i])
             continue;
 
-        const Item * item = items[i];
+        const Item * item = s_items[i];
 
         for (int entry : item->matches)
         {
             add.append (
-                aud_playlist_entry_get_filename (list, entry),
-                aud_playlist_entry_get_tuple (list, entry, Playlist::Nothing)
+                s_playlist.entry_filename (entry),
+                s_playlist.entry_tuple (entry, Playlist::NoWait),
+                s_playlist.entry_decoder (entry, Playlist::NoWait)
             );
         }
 
@@ -545,84 +546,95 @@ static void do_add (bool play, bool set_title)
             title = item->name;
     }
 
-    int list2 = aud_playlist_get_active ();
-    aud_playlist_entry_insert_batch (list2, -1, std::move (add), play);
+    auto list2 = Playlist::active_playlist ();
+    list2.insert_items (-1, std::move (add), play);
 
     if (set_title && n_selected == 1)
-        aud_playlist_set_title (list2, title);
+        list2.set_title (title);
 }
 
 static void action_play ()
 {
-    int list = aud_playlist_get_temporary ();
-    aud_playlist_set_active (list);
-
-    if (aud_get_bool (nullptr, "clear_playlist"))
-        aud_playlist_entry_delete (list, 0, aud_playlist_entry_count (list));
-    else
-        aud_playlist_queue_delete (list, 0, aud_playlist_queue_count (list));
-
+    Playlist::temporary_playlist ().activate ();
     do_add (true, false);
 }
 
 static void action_create_playlist ()
 {
-    aud_playlist_new ();
+    Playlist::new_playlist ();
     do_add (false, true);
 }
 
 static void action_add_to_playlist ()
 {
-    if (aud_playlist_by_unique_id (playlist_id) != aud_playlist_get_active ())
+    if (s_playlist != Playlist::active_playlist ())
         do_add (false, false);
 }
 
 static void list_get_value (void * user, int row, int column, GValue * value)
 {
-    g_return_if_fail (row >= 0 && row < items.len ());
+    static constexpr aud::array<SearchField, const char *> start_tags =
+        {"", "<b>", "<i>", ""};
+    static constexpr aud::array<SearchField, const char *> end_tags =
+        {"", "</b>", "</i>", ""};
 
-    const Item * item = items[row];
-    StringBuf string = str_concat ({item->name, "\n"});
+    auto escape = [] (const char * s)
+        { return CharPtr (g_markup_escape_text (s, -1)); };
+
+    g_return_if_fail (row >= 0 && row < s_items.len ());
+
+    const Item * item = s_items[row];
+
+    CharPtr name = escape ((item->field == SearchField::Genre) ?
+                           str_toupper_utf8 (item->name) : item->name);
+
+    StringBuf desc (0);
 
     if (item->field != SearchField::Title)
     {
-        string.insert (-1, " ");
-        string.combine (str_printf (dngettext (PACKAGE, "%d song", "%d songs",
+        desc.insert (-1, " ");
+        desc.combine (str_printf (dngettext (PACKAGE, "%d song", "%d songs",
          item->matches.len ()), item->matches.len ()));
     }
 
     if (item->field == SearchField::Genre)
     {
-        string.insert (-1, " ");
-        string.insert (-1, _("of this genre"));
+        desc.insert (-1, " ");
+        desc.insert (-1, _("of this genre"));
     }
 
-    while ((item = item->parent))
+    if (item->parent)
     {
-        string.insert (-1, " ");
-        string.insert (-1, (item->field == SearchField::Album) ? _("on") : _("by"));
-        string.insert (-1, " ");
-        string.insert (-1, item->name);
+        auto parent = (item->parent->parent ? item->parent->parent : item->parent);
+
+        desc.insert (-1, " ");
+        desc.insert (-1, (parent->field == SearchField::Album) ? _("on") : _("by"));
+        desc.insert (-1, " ");
+        desc.insert (-1, start_tags[parent->field]);
+        desc.insert (-1, escape (parent->name));
+        desc.insert (-1, end_tags[parent->field]);
     }
 
-    g_value_set_string (value, string);
+    g_value_take_string (value, g_strdup_printf
+     ("%s%s%s\n<small>%s</small>", start_tags[item->field], (const char *) name,
+      end_tags[item->field], (const char *) desc));
 }
 
 static bool list_get_selected (void * user, int row)
 {
-    g_return_val_if_fail (row >= 0 && row < selection.len (), false);
-    return selection[row];
+    g_return_val_if_fail (row >= 0 && row < s_selection.len (), false);
+    return s_selection[row];
 }
 
 static void list_set_selected (void * user, int row, bool selected)
 {
-    g_return_if_fail (row >= 0 && row < selection.len ());
-    selection[row] = selected;
+    g_return_if_fail (row >= 0 && row < s_selection.len ());
+    s_selection[row] = selected;
 }
 
 static void list_select_all (void * user, bool selected)
 {
-    for (bool & s : selection)
+    for (bool & s : s_selection)
         s = selected;
 }
 
@@ -647,30 +659,76 @@ static void list_right_click (void * user, GdkEventButton * event)
     gtk_menu_popup ((GtkMenu *) menu, nullptr, nullptr, nullptr, nullptr, event->button, event->time);
 }
 
+static Index<char> list_get_data (void * user)
+{
+    if (s_search_pending)
+        search_timeout ();
+
+    int n_items = s_items.len ();
+    Index<char> buf;
+
+    s_playlist.select_all (false);
+
+    for (int i = 0; i < n_items; i ++)
+    {
+        if (! s_selection[i])
+            continue;
+
+        const Item * item = s_items[i];
+
+        for (int entry : item->matches)
+        {
+            if (buf.len ())
+                buf.append ('\n');
+
+            String filename = s_playlist.entry_filename (entry);
+            buf.insert (filename, -1, strlen (filename));
+
+            s_playlist.select_entry (entry, true);
+        }
+    }
+
+    s_playlist.cache_selected ();
+
+    return buf;
+}
+
 static const AudguiListCallbacks list_callbacks = {
     list_get_value,
     list_get_selected,
     list_set_selected,
     list_select_all,
     list_activate_row,
-    list_right_click
+    list_right_click,
+    nullptr,  // shift_rows
+    "text/uri-list",
+    list_get_data,
+    nullptr  // receive_data
 };
 
 static void entry_cb (GtkEntry * entry, void * unused)
 {
     const char * text = gtk_entry_get_text ((GtkEntry *) entry);
-    search_terms = str_list_to_index (str_tolower_utf8 (text), " ");
-    search_timer.queue (SEARCH_DELAY, search_timeout, nullptr);
-    search_pending = true;
+    s_search_terms = str_list_to_index (str_tolower_utf8 (text), " ");
+    s_search_timer.queue (SEARCH_DELAY, search_timeout, nullptr);
+    s_search_pending = true;
 }
 
-static void refresh_cb (GtkButton * button, GtkWidget * chooser)
+static void file_entry_cb (GtkEntry * entry, GtkWidget * button)
 {
-    char * path = gtk_file_chooser_get_filename ((GtkFileChooser *) chooser);
-    begin_add (path);
-    g_free (path);
+    const char * text = gtk_entry_get_text ((GtkEntry *) entry);
+    gtk_widget_set_sensitive (button, text[0] != 0);
+}
 
-    update_database ();
+static void refresh_cb (GtkButton * button, GtkWidget * file_entry)
+{
+    String uri = audgui_file_entry_get_uri (file_entry);
+    if (uri)
+    {
+        audgui_file_entry_set_uri (file_entry, uri);  // normalize path
+        begin_add (uri);
+        update_database ();
+    }
 }
 
 void * SearchTool::get_gtk_widget ()
@@ -705,10 +763,10 @@ void * SearchTool::get_gtk_widget ()
     gtk_widget_set_no_show_all (scrolled, true);
     gtk_box_pack_start ((GtkBox *) vbox, scrolled, true, true, 0);
 
-    results_list = audgui_list_new (& list_callbacks, nullptr, items.len ());
+    results_list = audgui_list_new (& list_callbacks, nullptr, s_items.len ());
     g_signal_connect (results_list, "destroy", (GCallback) gtk_widget_destroyed, & results_list);
     gtk_tree_view_set_headers_visible ((GtkTreeView *) results_list, false);
-    audgui_list_add_column (results_list, nullptr, 0, G_TYPE_STRING, -1);
+    audgui_list_add_column (results_list, nullptr, 0, G_TYPE_STRING, -1, true);
     gtk_container_add ((GtkContainer *) scrolled, results_list);
 
     stats_label = gtk_label_new ("");
@@ -719,11 +777,11 @@ void * SearchTool::get_gtk_widget ()
     GtkWidget * hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 6);
     gtk_box_pack_end ((GtkBox *) vbox, hbox, false, false, 0);
 
-    GtkWidget * chooser = gtk_file_chooser_button_new (_("Choose Folder"),
-     GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
-    gtk_box_pack_start ((GtkBox *) hbox, chooser, true, true, 0);
+    GtkWidget * file_entry = audgui_file_entry_new
+     (GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER, _("Choose Folder"));
+    gtk_box_pack_start ((GtkBox *) hbox, file_entry, true, true, 0);
 
-    gtk_file_chooser_set_filename ((GtkFileChooser *) chooser, get_path ());
+    audgui_file_entry_set_uri (file_entry, get_uri ());
 
     GtkWidget * button = gtk_button_new ();
     gtk_container_add ((GtkContainer *) button, gtk_image_new_from_icon_name
@@ -736,7 +794,9 @@ void * SearchTool::get_gtk_widget ()
     g_signal_connect (vbox, "destroy", (GCallback) search_cleanup, nullptr);
     g_signal_connect (entry, "changed", (GCallback) entry_cb, nullptr);
     g_signal_connect (entry, "activate", (GCallback) action_play, nullptr);
-    g_signal_connect (button, "clicked", (GCallback) refresh_cb, chooser);
+    g_signal_connect (file_entry, "changed", (GCallback) file_entry_cb, button);
+    g_signal_connect (file_entry, "activate", (GCallback) refresh_cb, file_entry);
+    g_signal_connect (button, "clicked", (GCallback) refresh_cb, file_entry);
 
     gtk_widget_show_all (vbox);
     gtk_widget_show (results_list);
